"""Helper script to open napari with 2 test label layers with features"""
from pathlib import Path

import imageio
import napari

import numpy as np
from napari_feature_classifier.feature_loader_widget import (
    make_features,
)


def main():
    """Main function that opens napari with 2 test label layers with features"""
    lbls = imageio.v2.imread(Path("sample_data/test_labels.tif"))
    lbls2 = np.zeros_like(lbls)
    lbls2[:, 3:, 2:] = lbls[:, :-3, :-2]
    lbls2 = lbls2 * 20

    labels = np.unique(lbls)[1:]
    labels_2 = np.unique(lbls2)[1:]

    viewer = napari.Viewer()
    lbls_layer = viewer.add_labels(lbls)
    lbls_layer2 = viewer.add_labels(lbls2)

    lbls_layer.features = make_features(labels, roi_id="ROI1", n_features=6)
    lbls_layer2.features = make_features(labels_2, roi_id="ROI2", n_features=6)
    # classifier_widget = ClassifierWidget(viewer)

    # viewer.window.add_dock_widget(classifier_widget)
    viewer.show(block=True)


if __name__ == "__main__":
    main()



"""Classifier container widget for napari"""
import logging
import pickle

from packaging import version
from pathlib import Path
from typing import Optional

import napari
import napari.layers
import napari.viewer
import numpy as np
import pandas as pd
from magicgui.widgets import (
    Container,
    Label,
    FileEdit,
    RadioButtons,
    PushButton,
    Select,
)

from napari_feature_classifier.annotator_init_widget import LabelAnnotatorTextSelector
from napari_feature_classifier.annotator_widget import (
    LabelAnnotator,
    get_class_selection,
)
from napari_feature_classifier.classifier import Classifier
from napari_feature_classifier.utils import (
    get_colormap,
    reset_display_colormaps_modern,
    reset_display_colormaps_legacy,
    get_valid_label_layers,
    get_selected_or_valid_label_layer,
    napari_info,
    overwrite_check_passed,
    add_annotation_names,
    NapariHandler,
)


class ClassifierInitContainer(Container):
    """
    The ClassifierInitContainer presents all the options needed for
    initializing a `ClassifierRunContainer`. It's intended as a container
    that's used in other magicgui containers.

    It offers feature selection for the last label layer that was selected
    (or a valid label layer if none had been selected). Changing the selected
    label layer changes the feature selection that is offered.
    No action is bound to the _initialize_button => needs to be done in the
    parent container to bind the correct run action (e.g. start a
    `ClassifierRunContainer` with the correct parameters)

    Paramters
    ---------
    viewer: napari.Viewer
        The current napari.Viewer instance

    Attributes
    ----------
    viewer: napari.Viewer
        The current napari.Viewer instance
    _last_selected_label_layer: napari.layers.Labels
        The last selected label layer
    last_selected_layer_label: magicgui.widgets.Label
        The Label widget for displaying the last selected label layer
    _feature_combobox: magicgui.widgets.Select
        The Select widget for selecting the features to use for classification
    _annotation_name_selector: LabelAnnotatorTextSelector
        The LabelAnnotatorTextSelector widget for selecting the annotation names
    _initialize_button: magicgui.widgets.PushButton
        The PushButton widget for initializing the classifier
    """

    def __init__(self, viewer: napari.viewer.Viewer):
        self._viewer = viewer
        try:
            self._last_selected_label_layer = get_selected_or_valid_label_layer(
                viewer=self._viewer
            )
        except NotImplementedError:
            self._last_selected_label_layer = None
        # TODO: Make this label left-aligned, not centered
        self.last_selected_layer_label = Label(
            label="Selecting features from:", value=self._last_selected_label_layer
        )
        self._feature_combobox = Select(
            choices=self.get_feature_options(self._last_selected_label_layer),
            allow_multiple=True,
            label="Feature Selection:",
        )
        self._annotation_name_selector = LabelAnnotatorTextSelector()
        # pylint: disable=W0212
        self._initialize_button = PushButton(text="Initialize")
        super().__init__(
            widgets=[
                self.last_selected_layer_label,
                self._feature_combobox,
                self._annotation_name_selector,
                self._initialize_button,
            ]
        )
        self._viewer.layers.selection.events.changed.connect(
            self.update_layer_selection
        )

    def get_selected_features(self):
        """
        Returns the currently selected features (0-n features)
        """
        return self._feature_combobox.value

    def get_class_names(self):
        """
        Returns the available class names of the classifier
        """
        return self._annotation_name_selector.get_class_names()

    def get_feature_options(self, layer):
        """
        Get the feature options of the currently selected layer

        Only works if a label layer is selected (we don't load features from
        other layers)
        """
        if isinstance(layer, napari.layers.Labels):
            return list(layer.features.columns)
        return []

    def update_layer_selection(self):
        """
        Update the layer selection and the feature options if the newly
        selected layer is a label layer
        """
        if isinstance(self._viewer.layers.selection.active, napari.layers.Labels):
            self._last_selected_label_layer = self._viewer.layers.selection.active
            self.last_selected_layer_label.value = self._last_selected_label_layer
            self._feature_combobox.choices = self.get_feature_options(
                self._last_selected_label_layer
            )
            # pylint: disable=W0212
            self._feature_combobox._default_choices = self.get_feature_options(
                self._last_selected_label_layer
            )


# pylint: disable=R0902
class ClassifierRunContainer(Container):
    """
    The `ClassifierRunContainer` manages all the options needed to do
    annotations, train a classifier and show classifier results.

    The `ClassifierRunContainer` can be initialized with either an existing
    classifier or with class_names + feature_names

    Paramters
    ---------
    viewer: napari.Viewer
        The current napari.Viewer instance
    classifier: Optional[Classifier]
        The container can be initialized with an existing classifier. If none
        is provided, a classifier is generated upon init.
    class_names: Optional[list[str]]
        The class names of the classifier. Needs to be provided if no
        classifier is provided.
    feature_names: Optional[list[str]]
        The feature names of the classifier. Needs to be provided if no
        classifier is provided.
    classifier_save_path: Optional[str]
        The path to save the classifier to. If none is provided, the
        classifier creates a default path in the current working directory.
    auto_save: Optional[bool]
        Whether the classifier can automatically save at the
        classifier_save_path upon run and save. If false, run does not
        trigger a save and save checks for overwrite conflicts.

    Attributes
    ----------
    viewer: napari.Viewer
        The current napari.Viewer instance
    _last_selected_label_layer: napari.layers.Labels
        The last selected label layer
    _classifier: Classifier
        The classifier object
    class_names: list[str]
        The class names of the classifier classes. Matching the first name to
        1, second to 2 etc.
    feature_names: list[str]
        The feature names of the classifier. These features are loaded from
        layer.features to train the classifier
    _label_column: str
        The column name of the label column in the layer.features dataframe,
        hard-coded to "label"
    _roi_id_column: str
        The column name of the roi_id column in the layer.features dataframe,
        hard-coded to "roi_id"
    _annotator: LabelAnnotator
        The LabelAnnotator container that manages annotations
    _prediction_layer: napari.layers.Labels
        The prediction layer that is generated by the classifier and which
        displays predictions made for the currently selected label layer.
    _run_button: magicgui.widgets.PushButton
        The PushButton widget for running the classifier
    _save_destination: magicgui.widgets.FileEdit
        The FileEdit widget for selecting the save destination of the classifier
    _save_button: magicgui.widgets.PushButton
        The PushButton widget for saving the classifier
    """

    # pylint: disable=R0913
    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        classifier: Optional[Classifier] = None,
        class_names: Optional[list[str]] = None,
        feature_names: Optional[list[str]] = None,
        classifier_save_path: Optional[str] = None,
        auto_save: Optional[bool] = False,
    ):
        self._viewer = viewer
        self.auto_save = auto_save
        self._last_selected_label_layer = get_selected_or_valid_label_layer(
            viewer=self._viewer
        )

        # Initialize the classifier
        if classifier:
            self._classifier = classifier
            self.class_names = self._classifier.get_class_names()
            self.feature_names = self._classifier.get_feature_names()
        else:
            if not class_names or not feature_names:
                raise ValueError(
                    "A classifier object or "
                    "class_names & feature_names "
                    "must be provided"
                )
            self._classifier = Classifier(
                feature_names=feature_names, class_names=class_names
            )
            self.class_names = class_names
            self.feature_names = feature_names

        self._label_column = "label"
        self._roi_id_colum = "roi_id"

        self._annotator = LabelAnnotator(
            self._viewer, get_class_selection(class_names=self.class_names)
        )

        for layer in self._viewer.layers:
            if type(layer) == napari.layers.Labels and layer.name == "Predictions":
                self._viewer.layers.remove(layer)
        self.add_prediction_layer()

        # Set the label selection to a valid label layer => Running into proxy bug
        self._viewer.layers.selection.active = self._last_selected_label_layer

        self._run_button = PushButton(text="Run Classifier")
        self._save_destination = FileEdit(
            label="Classifier Save Path",
            value=f"{self._last_selected_label_layer}_classifier.clf",
            mode="w",
        )
        if classifier_save_path:
            self._save_destination.value = classifier_save_path
        self._save_button = PushButton(text="Save Classifier")

        # Export options
        self._export_destination = FileEdit(
            label="Prediction Export Path",
            value=f"{self._last_selected_label_layer}_predictions.csv",
            mode="w",
        )
        self._export_button = PushButton(text="Export Classifier Result")
        super().__init__(
            widgets=[
                self._annotator,
                self._save_destination,
                self._run_button,
                self._save_button,
                self._export_destination,
                self._export_button,
            ]
        )
        self._run_button.clicked.connect(self.run)
        self._save_button.clicked.connect(self.save)
        self._export_button.clicked.connect(self.export_results)
        self._viewer.layers.selection.events.changed.connect(self.selection_changed)
        self._init_prediction_layer(self._last_selected_label_layer)

    def run(self):
        """
        Run method that adds features to the classifier, trains it, triggers
        predictions & saves the classifier
        """

        self.add_features_to_classifier()
        try:
            self._classifier.train()
        except ValueError as e:
            napari_info(
                "Training failed. A typical reason are not having "
                "enough annotations. \nThe error message was: "
                f"{e}"
            )
        else:
            self.make_predictions()
            self._prediction_layer.visible = True
            self.save()

    def add_features_to_classifier(self):
        """
        Generate a dict of features: Key are roi_ids, values are dataframes
        from layer.features.
        """
        dict_of_features = {}
        for layer in self._viewer.layers:
            if (
                isinstance(layer, napari.layers.Labels)
                and len(layer.features) > 0
                and "annotations" in layer.features.columns
            ):
                # TODO: Add extra checks that it contains valid features?
                if "roi_id" in layer.features.columns:
                    roi_ids = layer.features["roi_id"].unique()
                    if len(roi_ids) > 1:
                        raise NotImplementedError(
                            f"{layer=} contained no-unique roi_ids: {roi_ids}"
                        )

                    roi_id = roi_ids[0]
                    dict_of_features[roi_id] = layer.features
                else:
                    # TODO: Consider label-layer hashing here instead of
                    # using the layer name as roi_id
                    dict_of_features[layer.name] = layer.features
        self._classifier.add_dict_of_features(dict_of_features)

    def add_prediction_layer(self):
        self._prediction_layer = self._viewer.add_labels(
            self._last_selected_label_layer.data,
            scale=self._last_selected_label_layer.scale,
            name="Predictions",
            translate=self._last_selected_label_layer.translate,
        )
        self._prediction_layer.contour = 2

    def make_predictions(self):
        """
        Make predictions for all relevant label layers and add them to the
        layer.features `prediction` column of each layer
        """
        # Get all the label layers that have fitting features
        relevant_label_layers = self.get_relevant_label_layers()

        # Get the features dataframes with the relevant features
        prediction_dfs = {}
        for label_layer in relevant_label_layers:
            roi_id = self.get_layer_roi_id(label_layer)
            if roi_id in prediction_dfs.keys():
                raise ValueError(
                    f"Duplicate roi_id {roi_id} found in {label_layer.name}. "
                    "It's already present as the roi_id of another label layer. "
                )
            prediction_dfs[roi_id] = self.get_relevant_features(
                label_layer.features, set_index=False
            )

        # Get the classifier predictions
        prediction_results_dict = self._classifier.predict_on_dict(prediction_dfs)

        # Append the predictions to each open label layer ("prediction" column)
        for label_layer in relevant_label_layers:
            roi_id = self.get_layer_roi_id(label_layer)
            if "prediction" in label_layer.features.columns:
                label_layer.features.drop(columns=["prediction"], inplace=True)
            # Merge the predictions back into the layer.features dataframe
            label_layer.features = label_layer.features.merge(
                prediction_results_dict[roi_id],
                left_on=[self._roi_id_colum, self._label_column],
                right_index=True,
                how="outer",
            )

        self._init_prediction_layer(self._last_selected_label_layer)

    def selection_changed(self):
        """
        Check if the selection change results in a valid label layer being
        selected. If so, initialize the prediction layer for it.
        """
        if self._viewer.layers.selection.active:
            if self._viewer.layers.selection.active in get_valid_label_layers(
                viewer=self._viewer
            ):
                self._last_selected_label_layer = self._viewer.layers.selection.active
                self._init_prediction_layer(
                    self._viewer.layers.selection.active, ensure_layer_presence=False
                )
                self._update_export_destination(self._last_selected_label_layer)

    def reorder_layers(self):
        """Reorders layers if needed to ensure Annotation & Prediction layers
        are above the currently selected label layer.
        """
        # Get the current order of layers
        all_layers = list(self._viewer.layers)

        # Determine the indices of the layers if they exist
        indices_to_move = []

        # Find the index of "Prediction" layer if it exists
        if "Predictions" in self._viewer.layers:
            indices_to_move.append(self._viewer.layers.index("Predictions"))

        # Find the index of "Annotation" layer if it exists
        if "Annotations" in self._viewer.layers:
            indices_to_move.append(self._viewer.layers.index("Annotations"))

        # Find the index of the reference_label_layer
        if self._last_selected_label_layer.name in self._viewer.layers:
            indices_to_move.append(
                self._viewer.layers.index(self._last_selected_label_layer.name)
            )

        # Calculate the new order of layer indices
        remaining_indices = [
            i for i in range(len(all_layers)) if i not in indices_to_move
        ]
        remaining_indices.reverse()
        new_order = indices_to_move + remaining_indices
        new_order.reverse()

        # Reorder the layers using the move_multiple function
        self._viewer.layers.move_multiple(new_order)

    def _init_prediction_layer(
        self, label_layer: napari.layers.Labels, ensure_layer_presence: bool = True
    ):
        """
        Initialize the prediction layer and reset its data (to fit the input
        label_layer) and its colormap.
        ensure_layer_presence creates the Predictions layer if it doesn't exist
        yet and triggers layer reordering.
        """
        # Ensure that prediction layer exists
        if (
            "Predictions" not in [x.name for x in self._viewer.layers]
            and ensure_layer_presence
        ):
            self.add_prediction_layer()
        if ensure_layer_presence:
            # Ensure correct layer order: This sometimes fails with weird
            # EmitLoopError & IndexError that should be ignored
            try:
                self.reorder_layers()
            except:  # noqa
                pass

        # Ensure that prediction layer is above the current label layer
        self._last_selected_label_layer

        # Check if the predict column already exists in the layer.features
        if "prediction" not in label_layer.features:
            unique_labels = np.unique(label_layer.data)[1:]
            predict_df = pd.DataFrame(
                {self._label_column: unique_labels, "prediction": np.NaN}
            )
            if self._label_column in label_layer.features.columns:
                label_layer.features = label_layer.features.merge(
                    predict_df, on=self._label_column, how="outer"
                )
            else:
                label_layer.features = pd.concat(
                    [label_layer.features, predict_df], axis=1
                )

        # Update the label data in the prediction layer
        self._prediction_layer.data = label_layer.data
        self._prediction_layer.scale = label_layer.scale
        self._prediction_layer.translate = label_layer.translate

        # Update the colormap of the prediction layer
        napari_version = version.parse(napari.__version__)
        if napari_version >= version.parse("0.4.19"):
            reset_display_colormaps_modern(
                label_layer,
                feature_col="prediction",
                display_layer=self._prediction_layer,
                label_column=self._label_column,
                cmap=get_colormap(),
            )
        else:
            reset_display_colormaps_legacy(
                label_layer,
                feature_col="prediction",
                display_layer=self._prediction_layer,
                label_column=self._label_column,
                cmap=get_colormap(),
            )

    def get_relevant_label_layers(self):
        relevant_label_layers = []
        required_columns = [self._label_column, self._roi_id_colum]
        excluded_label_layers = ["Annotations", "Predictions"]
        for label_layer in self._viewer.layers:
            if (
                isinstance(label_layer, napari.layers.Labels)
                and label_layer.name not in excluded_label_layers
            ):
                if label_layer.features is not None:
                    if all(x in label_layer.features.columns for x in required_columns):
                        relevant_label_layers.append(label_layer)
        return relevant_label_layers

    def get_layer_roi_id(self, label_layer):
        roi_ids = label_layer.features[self._roi_id_colum].unique()
        if len(roi_ids) > 1:
            raise NotImplementedError(
                f"{label_layer=} contained no-unique roi_ids: {roi_ids}"
            )
        return roi_ids[0]

    # pylint: disable=C0103
    def get_relevant_features(
        self, df, filter_annotations: bool = False, set_index=False
    ):
        """
        Get the relevant features from the pandas table
        Can optionally create a double-indexing with label & roi_id
        filter_annotations: Only return rows that contain annotations?
        """
        if not filter_annotations:
            df_relevant = df[
                [*self.feature_names, self._label_column, self._roi_id_colum]
            ]
        else:
            df_relevant = df.loc[
                df["annotations"].notna(),
                [
                    *self.feature_names,
                    self._label_column,
                    self._roi_id_colum,
                    "annotations",
                ],
            ]
        if set_index:
            df_relevant.set_index(
                [self._roi_id_colum, self._label_column], inplace=True
            )
        return df_relevant

    def save(self):
        """
        Save the classifier and handle overwriting of existing classifier file
        """
        if not self.auto_save:
            # Handle existing classifier file => ask for overwrite
            if not overwrite_check_passed(
                file_path=self._save_destination.value, output_type="classifier"
            ):
                return
        # If the user confirms overwriting the classifier once, keep
        # overwriting it going forward. We want classifier auto-save, just not
        # overwriting of other existing classifiers with the same name.
        self.auto_save = True
        output_path = Path(self._save_destination.value)
        self._classifier.save(output_path)

    def _update_export_destination(self, label_layer: napari.layers.Labels):
        """
        Update the default export destination to the name of the label layer.
        If a base_path was already set, keep it on that base path.

        """
        base_path = Path(self._export_destination.value).parent
        self._export_destination.value = (
            base_path / f"{label_layer.name}_predictions.csv"
        )

    def export_results(self):
        """
        Export classifier results for the current layer if available
        """
        if not overwrite_check_passed(
            file_path=self._export_destination.value, output_type="predictions"
        ):
            return

        predictions = self._last_selected_label_layer.features.loc[
            :, [self._label_column, "prediction", "annotations"]
        ]
        # pylint: disable=C0103
        df = add_annotation_names(
            df=pd.DataFrame(predictions), ClassSelection=self._annotator.ClassSelection
        )

        df.to_csv(self._export_destination.value)
        napari_info(f"Annotations were saved at {self._export_destination.value}")


class LoadClassifierContainer(Container):
    """
    The `LoadClassifierContainer` is a second entry-way to the classifier and
    can launch an appropriate `ClassifierRunContainer`.

    Paramters
    ---------
    viewer: napari.Viewer
        The current napari.Viewer instance

    Attributes
    ----------
    viewer: napari.Viewer
        The current napari.Viewer instance
    _clf_destination: magicgui.widgets.FileEdit
        The file edit widget that allows the user to select a classifier file
    _load_button: magicgui.widgets.PushButton
        The button that launches the `ClassifierRunContainer`
    _filter: magicgui.widgets.RadioButtons
        The radio button widget that allows the user to select the file filter
        to use for selecting the classifier file. See
        https://github.com/fractal-napari-plugins-collection/napari-feature-classifier/issues/36
        for more details.
    _run_container: ClassifierRunContainer
         The `ClassifierRunContainer` that is launched by the `_load_button`
    """

    def __init__(self, viewer: napari.viewer.Viewer):
        self._viewer = viewer
        self._clf_destination = FileEdit(mode="r", filter=None)
        self._filter = RadioButtons(
            value="*.clf",
            choices=["*.clf", "*.pkl", "*"],
            orientation="horizontal",
            label="Filter",
        )
        self._load_button = PushButton(label="Load Classifier")
        self._run_container = None
        super().__init__(
            widgets=[self._clf_destination, self._filter, self._load_button]
        )
        self._load_button.clicked.connect(self.load)
        self._filter.changed.connect(self.set_filter)

    def set_filter(self):
        """
        Updates the filter that is applied to the file edit widget
        """
        self._clf_destination.filter = self._filter.value

    def load(self):
        """
        Load a classifier from a file and start the run container with the
        correct options(already set classifier_save_path and turn on auto_save)
        """
        clf_path = Path(self._clf_destination.value)
        with open(clf_path, "rb") as f:  # pylint: disable=C0103
            clf = pickle.load(f)

        try:
            self._run_container = ClassifierRunContainer(
                self._viewer,
                clf,
                classifier_save_path=clf_path,
                auto_save=True,
            )
        except NotImplementedError:
            napari_info(
                "Create a label layer with a feature dataframe before loading "
                "the classifier"
            )
            return
        self.clear()
        self.append(self._run_container)

        # TODO: Add functionality that loads existing annotations from the
        # classifier and adds them back to the currently open label images


class ClassifierWidget(Container):
    """
    The `ClassifierWidget` is the parent widget and the one that is added as
    a dockwidget. It manages the `ClassifierInitContainer` and the
    `ClassifierRunContainer`.

    Paramters
    ---------
    viewer: napari.Viewer
        The current napari.Viewer instance

    Attributes
    ----------
    viewer: napari.Viewer
        The current napari.Viewer instance
    _run_container: None or ClassifierRunContainer
        The `ClassifierRunContainer` that's started.
    _init_container: None or ClassifierInitContainer
        The `ClassifierInitContainer` that's started.
    """

    def __init__(self, viewer: napari.viewer.Viewer):
        self._viewer = viewer

        self._init_container = None
        self._run_container = None
        self._init_container = None
        self.setup_logging()

        super().__init__(widgets=[])

        self.initialize_init_widget()

    def setup_logging(self):
        # Create a custom handler for napari
        napari_handler = NapariHandler()
        napari_handler.setLevel(logging.INFO)

        # Optionally, set a formatter for the handler
        # formatter = logging.Formatter(
        #     '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        # )
        # napari_handler.setFormatter(formatter)

        # Get the classifier's logger and add the napari handler to it
        classifier_logger = logging.getLogger("classifier")
        classifier_logger.addHandler(napari_handler)

    def initialize_init_widget(self):
        self._init_container = ClassifierInitContainer(self._viewer)
        self.append(self._init_container)
        self._init_container._initialize_button.clicked.connect(
            self.initialize_run_widget
        )

    def initialize_run_widget(self):
        class_names = self._init_container.get_class_names()
        feature_names = self._init_container.get_selected_features()
        if not feature_names:
            napari_info("No features selected")
            return
        self._run_container = ClassifierRunContainer(
            self._viewer, class_names=class_names, feature_names=feature_names
        )
        self.clear()
        self.append(self._run_container)



"""Widget for loading features from csv file."""
from os import PathLike
from typing import Callable, Sequence, Union
from typing_extensions import TypeAlias
import warnings

import numpy as np
import pandas as pd
import pandera as pa
from magicgui import magic_factory
from napari.layers import Labels
from napari.types import LayerDataTuple
from pandera.typing import DataFrame, Series
from pathlib import Path

from napari_feature_classifier.utils import napari_info

# pandera.SchemaModel is now deprecated 
# see https://github.com/unionai-oss/pandera/releases/tag/v0.20.0
class LabelFeatureSchema(pa.DataFrameModel):
    # roi_id: Series[str] = pa.Field(coerce=True, unique=False)
    label: Series[int] = pa.Field(coerce=True, unique=True)


@pa.check_types
def load_features_csv(
    fn: PathLike, index_column_or_columns: Union[str, list[str]] = "label"
) -> DataFrame[LabelFeatureSchema]:
    df = pd.read_csv(fn)
    if isinstance(index_column_or_columns, str):
        assert (
            index_column_or_columns in df
        ), f"missing index column `{index_column_or_columns}` in csv file."
        return DataFrame[LabelFeatureSchema](
            df.rename(columns={index_column_or_columns: "label"})
        )
    # return pd.DataFrame(data=np.random.rand(5, 5), columns=list("abcde"),
    # index=list(range(1, 10, 2)))
    return DataFrame[LabelFeatureSchema](df)


@pa.check_types
def make_features(
    labels: Sequence[int], roi_id: str = "id", n_features: int = 10, seed: int = 42
) -> DataFrame[LabelFeatureSchema]:
    columns = [f"feature_{i}" for i in range(n_features)]
    rng = np.random.default_rng(seed=seed)
    features = rng.random(size=(len(labels), n_features))
    data = {
        **{"roi_id": roi_id, "label": labels},
        **{column: feature for column, feature in zip(columns, features.T)},
    }
    return DataFrame[LabelFeatureSchema](data)


FeatureLoaderFn: TypeAlias = Callable[[PathLike[str]], DataFrame[LabelFeatureSchema]]


@magic_factory(call_button="load features")
def load_features_factory(
    layer: Labels, path: Path, loader: FeatureLoaderFn = load_features_csv
) -> LayerDataTuple:
    df = loader(path) # pylint: disable=C0103
    image_labels = np.unique(layer.data)[1:]
    feature_labels = df["label"].values
    if len(set(image_labels).symmetric_difference(feature_labels)) != 0:
        warn_str = "Label image labels do not match with feature table.\n"
        "Label objects with no features: "
        f"{sorted(set(image_labels).difference(feature_labels))}\n"
        "Features with no label objects: "
        f"{sorted(set(feature_labels).difference(image_labels))}"
        napari_info(warn_str)
        warnings.warn(warn_str)
    napari_info(f"Loaded features and attached them to \"{layer}\" layer")
    return (layer.data, {"name": layer.name, "features": df}, "labels")



"""Annotator init widget for napari"""
import napari
import napari.layers
import napari.viewer
from magicgui.widgets import Container, LineEdit, PushButton

from napari_feature_classifier.annotator_widget import (
    LabelAnnotator,
    get_class_selection,
)


class LabelAnnotatorTextSelector(Container):
    """
    The `LabelAnnotatorTextSelector` container is a helper container for the
    label annotator where the user can name the classes to be annotated.

    Paramters
    ---------
    default_n_classes: int
        The number of classes to display. Defaults to 5.
    """

    MAX_CLASSES: int = 9

    def __init__(self, default_n_classes=5):
        default_line_edits = [
            LineEdit(value=f"Class_{i + 1}", nullable=True)
            for i in range(default_n_classes)
        ]
        empty_line_edits = [
            LineEdit(nullable=True) for i in range(self.MAX_CLASSES - default_n_classes)
        ]

        self._text_edits = tuple([*default_line_edits, *empty_line_edits])

        super().__init__(widgets=[*self._text_edits])

    def get_class_names(self):
        class_names = [e.value for e in self._text_edits if e.value != ""]
        return class_names


class InitializeLabelAnnotatorWidget(Container):
    """
    The `InitializeLabelAnnotatorWidget` container is an entry point to start
    an annotator without a classifier.

    Paramters
    ---------
    viewer: napari.Viewer
        The current napari.Viewer instance
    default_n_classes: int
        The number of classes to display. Defaults to 5.
    """

    def __init__(self, viewer: napari.viewer.Viewer, default_n_classes=5):
        self.viewer = viewer
        self.label_class_container = LabelAnnotatorTextSelector(default_n_classes)
        self._init_button = PushButton(label="Initialize")
        super().__init__(widgets=[self.label_class_container, self._init_button])
        self._init_button.clicked.connect(self.initialize_annotator)

    def initialize_annotator(self):
        class_names = self.label_class_container.get_class_names()
        annotator = LabelAnnotator(
            self.viewer, get_class_selection(class_names=class_names)
        )
        self.clear()
        self.append(annotator)



"""Utils function for the classifier"""
from functools import lru_cache
import logging
import math
from pathlib import Path

# import warnings
import pandas as pd
from napari.utils.notifications import show_info
from matplotlib.colors import ListedColormap
import matplotlib
import numpy as np
import napari
from qtpy.QtWidgets import QMessageBox  # pylint: disable=E0611

# from napari._qt.dialogs.qt_notification import NapariQtNotification
# from napari._qt.qt_event_loop import _ipython_has_eventloop


@lru_cache(maxsize=16)
def get_df(path):
    """
    Pandas csv reader function with caching

    Parameters
    ----------
    path: str or Path
        Path to the csv file to be loaded
    """
    return pd.read_csv(path)


def in_notebook():
    """
    Checks whether the plugin is run from within a jupyter notebook

    Returns
    -------
    boolean
        True if it's running in a jupyter notebook
    """

    # Check if I'm running in jupyter notebook, from here:
    # https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    try:
        from IPython import get_ipython  # pylint: disable-msg=C0415

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def get_colormap(matplotlib_colormap="Set1"):
    """
    Generates colormaps depending on the number of classes
    """
    new_colors = np.array(matplotlib.colormaps[matplotlib_colormap].colors).astype(
        np.float32
    )
    cmap_np = np.zeros(
        shape=(new_colors.shape[0] + 1, new_colors.shape[1] + 1), dtype=np.float32
    )
    cmap_np[1:, :-1] = new_colors
    cmap_np[1:, -1] = 1
    cmap = ListedColormap(cmap_np)
    return cmap


def reset_display_colormaps_legacy(
    label_layer, feature_col, display_layer, label_column, cmap
):
    """
    Reset the colormap based on the annotations in
    label_layer.features['annotation'] and sends the updated colormap
    to the annotation label layer
    """
    colors = cmap(label_layer.features[feature_col].astype(float) / len(cmap.colors))
    colordict = dict(zip(label_layer.features[label_column], colors))
    display_layer.color = colordict
    display_layer.opacity = 1.0
    display_layer.color_mode = "direct"


def reset_display_colormaps_modern(
    label_layer, feature_col, display_layer, label_column, cmap
):
    """
    Reset the colormap based on the annotations in
    label_layer.features['annotation'] and sends the updated colormap
    to the annotation label layer

    Modern version to support napari >= 0.4.19
    """
    from napari.utils.colormaps import DirectLabelColormap

    colors = cmap(label_layer.features[feature_col].astype(float) / len(cmap.colors))
    colordict = dict(zip(label_layer.features[label_column], colors))
    colordict[None] = [0, 0, 0, 0]
    display_layer.colormap = DirectLabelColormap(color_dict=colordict)
    display_layer.opacity = 1.0


def napari_info(message):
    """
    Info message wrapper.
    Ensures info is shown in napari (when napari is run from the command line)
    or printed (when napari is run from a jupyter notebook)
    If napari show_info can't be called (e.g. napari isn't running),
    it's skipped

    message
    ----------
    path: str
        Message to be shown to the user
    """
    try:
        show_info(message)
    except:  # pylint: disable=bare-except # noqa #E722
        print(message)
    # TODO: Would be better to check if it's running in napari and print in all
    # other cases (e.g. if someone runs the classifier form a script).
    # But can't make that work at the moment
    if in_notebook():
        print(message)


class NapariHandler(logging.Handler):
    def emit(self, record):
        log_entry = self.format(record)
        napari_info(log_entry)


def get_valid_label_layers(viewer) -> list[str]:
    """
    Get a list of label layers that are not `Annotations` or `Predictions`.
    """
    return [
        layer
        for layer in viewer.layers
        if isinstance(layer, napari.layers.Labels)
        and layer.name not in ["Annotations", "Predictions"]
    ]


def get_selected_or_valid_label_layer(viewer) -> napari.layers.Labels:
    """
    Get the selected label layer, or the first valid label layer.
    This is None if no layer or multiple layers are selected.
    """
    selected_layer = viewer.layers.selection.active
    valid_layers = get_valid_label_layers(viewer=viewer)
    if selected_layer and selected_layer in valid_layers:
        return viewer.layers[selected_layer.name]
    if len(valid_layers) > 0:
        return valid_layers[0]
    raise NotImplementedError("No valid label layers were found")


def overwrite_check_passed(file_path, output_type: str = ""):
    """
    If a file already exists, ask whether it should be overwritten.
    """
    if Path(file_path).exists():
        msg_box = QMessageBox()
        msg_box.setText(
            "Do you you want to overwrite the "
            f"existing {output_type}: "
            f"{file_path}?"
        )
        msg_box.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        msg_box.setDefaultButton(QMessageBox.Yes)

        response = msg_box.exec_()
        if not response == QMessageBox.Yes:
            return False
    return True


# pylint: disable=C0103
def add_annotation_names(df, ClassSelection):
    """
    Add a column with the actual annotation names to the dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with annotations column.
    ClassSelection : Enum
        Enum with the class names.

    Returns
    -------
    pd.DataFrame
    """
    class_names = []
    for annotation in df["annotations"]:
        if math.isnan(annotation):
            class_names.append(np.NaN)
        else:
            class_names.append(ClassSelection(annotation).name)
    df["annotation_names"] = class_names
    return df



"""Annotator container widget for napari"""
import warnings
from enum import Enum
from functools import partial
from packaging import version
from pathlib import Path
from typing import Optional, Sequence, cast

# pylint: disable=R0801
import napari
import napari.layers
import napari.viewer
import numpy as np
import pandas as pd
from magicgui.widgets import (
    Label,
    Container,
    FileEdit,
    PushButton,
    RadioButtons,
    create_widget,
)


# pylint: disable=R0801
from napari_feature_classifier.utils import (
    get_colormap,
    reset_display_colormaps_legacy,
    reset_display_colormaps_modern,
    get_valid_label_layers,
    get_selected_or_valid_label_layer,
    napari_info,
    overwrite_check_passed,
    add_annotation_names,
)


def get_class_selection(
    n_classes: Optional[int] = None, class_names: Optional[Sequence[str]] = None
) -> Enum:
    """
    Create a class selection enum for the annotator widget.

    Parameters:
    -----------
    n_classes: int, optional
        Number of classes to create the enum for. If not provided, the length of
        `class_names` is used.
    class_names: Sequence[str], optional
        List of class names to create the enum for. If not provided, the enum
        values will be `Class_1`, `Class_2`, etc., based on n_classes
    """
    if n_classes is None and class_names is None:
        raise ValueError("Provide either `n_classes` or a list of `class_names`")
    if class_names is None:
        class_names = [f"Class_{i+1}" for i in range(n_classes)]
    if n_classes is None:
        n_classes = len(class_names)
    if n_classes != len(class_names):
        warnings.warn(
            f"Value provided for `n_classes` ({n_classes}) does not match "
            f"the length of `class_names` ({len(class_names)}). "
            f"Setting n_classes to {len(class_names)}."
        )
    assert len(class_names) == len(
        set(class_names)
    ), f"{class_names=} contains duplicate entries!"

    # Setting NoClass to -1.0 => Deselection that is sent to the classifier
    # pylint: disable=C0103
    ClassSelection = Enum(
        "ClassSelection",
        {"NoClass": -1.0, **{c: i + 1 for i, c in enumerate(class_names)}},
    )
    return ClassSelection


# pylint: disable=R0902
class LabelAnnotator(Container):
    """
    The `LabelAnnotator` widget manages the annotation of a label layer by
    monitoring clicks on the selected label layer, adding annotations to the
    layer.features df and coloring an annotation layer accordingly.

    Paramters
    ---------
    viewer: napari.Viewer
        The current napari.Viewer instance
    ClassSelection: Enum
        The class selection to use for the annotation. Defaults to a 4 class selection.

    Attributes
    ----------
    viewer: napari.Viewer
        The current napari.Viewer instance
    _label_column: str
        The column name of the label column in the layer.features dataframe,
        hard-coded to "label"
    _last_selected_label_layer: napari.layers.Labels
        The last selected valid label layer
    last_selected_layer_label: magicgui.widgets.Label
        The Label widget for displaying the last selected label layer
    _annotations_layer: napari.layers.Labels
        The layer to on which annotations are displayed. This layer is not
        editable by the user.
    ClassSelection: Enum
        The class selection to use for the annotation.
    nb_classes: int
        The number of classes in the class selection (not counting deselection)
    cmap: matplotlib.colors.Colormap
        The colormap to use for the annotation layer
    _class_selector: magicgui.widgets.RadioButtons
        The RadioButtons widget for selecting the class to annotate.
        Can also be controlled via the number keys.
    """

    # TODO: Do we need to keep the annotation layer on top when new
    # annotations are made?
    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        ClassSelection=get_class_selection(n_classes=4),
    ):
        self._viewer = viewer
        self._label_column = "label"

        self._last_selected_label_layer = get_selected_or_valid_label_layer(
            viewer=self._viewer
        )

        self.last_selected_layer_label = Label(
            label="Last selected label layer:", value=self._last_selected_label_layer
        )

        # Handle existing predictions layer
        for layer in self._viewer.layers:
            if type(layer) == napari.layers.Labels and layer.name == "Annotations":
                self._viewer.layers.remove(layer)
        self.add_annotations_layer()

        # Class selection
        self.ClassSelection = ClassSelection  # pylint: disable=C0103
        self.nb_classes = len(self.ClassSelection) - 1
        self.cmap = get_colormap()
        self._class_selector = cast(
            RadioButtons,
            create_widget(
                label="Class Selection",
                value=ClassSelection[list(ClassSelection.__members__.keys())[1]],
                annotation=ClassSelection,
                widget_type=RadioButtons,
            ),
        )
        self._init_annotation(self._last_selected_label_layer)
        self._save_destination = FileEdit(
            label="Save Path", value="annotation.csv", mode="w"
        )
        self._save_annotation = PushButton(label="Save Annotations")
        self._update_save_destination(self._last_selected_label_layer)
        super().__init__(
            widgets=[
                self.last_selected_layer_label,
                self._class_selector,
                self._save_destination,
                self._save_annotation,
            ]
        )
        self._save_annotation.clicked.connect(self._on_save_clicked)
        # TODO: Connect to the user clicking save in the file dialog. I can
        # trigger an event that the user clicked the button to open the
        # file dialog (see below), but don't get the info whether the user
        # clicked confirm or cancel.
        # self._save_destination.choose_btn.changed.connect(self.user_accepted)
        # Connect to label layer change, potentially call init
        self._viewer.layers.selection.events.changed.connect(self.selection_changed)

    def selection_changed(self, event):
        """
        Callback for when the selection changes. If the selection change results
        in a valid label layer being selected, initialize the annotator for it.
        """
        # Check if the selection change results in a valid label layer being
        # selected. If so, initialize the annotator for it.
        if self._viewer.layers.selection.active:
            if self._viewer.layers.selection.active in get_valid_label_layers(
                viewer=self._viewer
            ):
                self._init_annotation(self._viewer.layers.selection.active)
                self._save_annotation.enabled = True
                self._save_destination.enabled = True
                self._class_selector.enabled = True
                self.last_selected_layer_label.value = (
                    self._viewer.layers.selection.active
                )
                self._last_selected_label_layer = self._viewer.layers.selection.active
                self._update_save_destination(self._last_selected_label_layer)
            else:
                self._save_annotation.enabled = False
                self._save_destination.enabled = False
                self._class_selector.enabled = False
        else:
            self._save_annotation.enabled = False
            self._save_destination.enabled = False
            self._class_selector.enabled = False

    def add_annotations_layer(self):
        self._annotations_layer = self._viewer.add_labels(
            self._last_selected_label_layer.data,
            scale=self._last_selected_label_layer.scale,
            name="Annotations",
            translate=self._last_selected_label_layer.translate,
        )
        self._annotations_layer.editable = False
        # Set the label selection to a valid label layer
        self._viewer.layers.selection.active = self._last_selected_label_layer

    def toggle_label(self, labels_layer, event):
        """
        Callback for when a label is clicked. It then updates the color of that
        label in the annotation layer.
        """
        # If the annotations layer is missing, add it back
        if "Annotations" not in [x.name for x in self._viewer.layers]:
            self.add_annotations_layer()

        scaled_position = self.get_scaled_position(
            event.position, labels_layer.translate, labels_layer.scale
        )
        label = labels_layer.get_value(scaled_position)
        if label == 0 or not label:
            napari_info(f"No label clicked on the {labels_layer} label layer.")
            return

        # Left click: add annotation
        if event.button == 1:
            labels_layer.features.loc[
                labels_layer.features[self._label_column] == label, "annotations"
            ] = self._class_selector.value.value
        # Right click: Remove annotation
        elif event.button == 2:
            labels_layer.features.loc[
                labels_layer.features[self._label_column] == label, "annotations"
            ] = np.NaN

        # Update only the single color value that changed
        napari_version = version.parse(napari.__version__)
        if napari_version >= version.parse("0.4.19"):
            self.update_single_color_slow(labels_layer, label)
        else:
            self.update_single_color_legacy(labels_layer, label)

    @staticmethod
    def get_scaled_position(
        position: tuple, translate: np.array, scale: np.array
    ) -> tuple:
        """
        Get the position of a click after translation & scaling


        Position values in napari can have different shapes than the layer
        translate & scale data (e.g. 3D position for 2D layer). This function
        handles that edge-case
        """
        if len(position) == 3 and len(scale) == 2:
            position = position[1:]
        elif len(position) != len(scale):
            raise NotImplementedError(
                "Detecting annotation positions isn't implemented for "
                f"positions like {position} of length {len(position)} and "
                f"scales like {scale} of length {len(scale)}"
            )
        return tuple(
            (pos - trans) / scale
            for pos, trans, scale in zip(position, translate, scale)
        )

    def set_class_n(self, event, n: int):  # pylint: disable=C0103
        self._class_selector.value = self.ClassSelection[
            list(self.ClassSelection.__members__)[n]
        ]

    def _init_annotation(self, label_layer: napari.layers.Labels):
        """
        Initializes the annotation layer for the given label layer.
        """
        label_layer.editable = False
        if "annotations" not in label_layer.features:
            unique_labels = np.unique(label_layer.data)[1:]
            annotation_df = pd.DataFrame(
                {self._label_column: unique_labels, "annotations": np.NaN}
            )
            if self._label_column in label_layer.features.columns:
                label_layer.features = label_layer.features.merge(
                    annotation_df, on=self._label_column, how="outer"
                )
            else:
                label_layer.features = pd.concat(
                    [label_layer.features, annotation_df], axis=1
                )

        self._annotations_layer.data = label_layer.data
        self._annotations_layer.scale = label_layer.scale
        self._annotations_layer.translate = label_layer.translate

        napari_version = version.parse(napari.__version__)
        if napari_version >= version.parse("0.4.19"):
            reset_display_colormaps_modern(
                label_layer,
                feature_col="annotations",
                display_layer=self._annotations_layer,
                label_column=self._label_column,
                cmap=self.cmap,
            )
        else:
            reset_display_colormaps_legacy(
                label_layer,
                feature_col="annotations",
                display_layer=self._annotations_layer,
                label_column=self._label_column,
                cmap=self.cmap,
            )
        if self.toggle_label not in label_layer.mouse_drag_callbacks:
            label_layer.mouse_drag_callbacks.append(self.toggle_label)

        # keybindings for the available classes (0 = deselect)
        for i in range(len(self.ClassSelection)):
            set_class = partial(self.set_class_n, n=i)
            set_class.__name__ = f"set_class_{i}"
            label_layer.bind_key(str(i), set_class, overwrite=True)

    def _update_save_destination(self, label_layer: napari.layers.Labels):
        """
        Update the default save destination to the name of the label layer.
        If a base_path was already set, keep it on that base path.

        """
        base_path = Path(self._save_destination.value).parent
        self._save_destination.value = base_path / f"{label_layer.name}_annotation.csv"

    def update_single_color_legacy(self, label_layer, label):
        """
        Update the color of a single object in the annotations layer.
        """
        color = self.cmap(
            float(
                label_layer.features.loc[
                    label_layer.features[self._label_column] == label,
                    "annotations",
                ].iloc[0]
            )
            / len(self.cmap.colors)
        )
        self._annotations_layer.color[label] = color
        self._annotations_layer.opacity = 1.0
        self._annotations_layer.color_mode = "direct"

    def update_single_color_slow(self, label_layer, label):
        """
        Update the color of a single object in the annotations layer.

        napari >= 0.4.19 does not have a direct API to only update a single
        color. It always validates & updates the whole colormap.
        Therefore, this update mode scales badly with the number of unique
        labels.
        See details in https://github.com/napari/napari/issues/6732

        """
        color = self.cmap(
            float(
                label_layer.features.loc[
                    label_layer.features[self._label_column] == label,
                    "annotations",
                ].iloc[0]
            )
            / len(self.cmap.colors)
        )
        from napari.utils.colormaps import DirectLabelColormap

        colordict = self._annotations_layer.colormap.color_dict
        colordict[label] = color
        self._annotations_layer.colormap = DirectLabelColormap(color_dict=colordict)
        self._annotations_layer.opacity = 1.0

    def _on_save_clicked(self):
        """
        Save annotations to a csv file.
        """
        # Check whether annotations should be overwritten.
        if not overwrite_check_passed(
            file_path=self._save_destination.value, output_type="annotation export"
        ):
            return

        annotations = self._last_selected_label_layer.features.loc[
            :, [self._label_column, "annotations"]
        ]
        # pylint: disable=C0103
        df = add_annotation_names(
            df=pd.DataFrame(annotations), ClassSelection=self.ClassSelection
        )

        df.to_csv(self._save_destination.value)
        napari_info(f"Annotations were saved at {self._save_destination.value}")



# pylint: disable=C0103
"""Tests for core classifier class"""
import numpy as np
import pandas as pd
import pytest

from napari_feature_classifier.classifier import Classifier

# pylint: disable=C0103
CLASSIFIER_FEATURE_NAMES = ["feature1", "feature2", "feature3"]
DROP_COLUMNS = ["feature6", "feature7"]
TABLE_FEATURE_NAMES = CLASSIFIER_FEATURE_NAMES + DROP_COLUMNS
TABLE_FEATURES = {
    feature_name: np.random.randn(5) for feature_name in TABLE_FEATURE_NAMES
}
CLASS_NAMES = ["s", "m"]
INDEX_COLUMNS = ["roi_id", "label"]


def get_classifier():
    return Classifier(feature_names=CLASSIFIER_FEATURE_NAMES, class_names=CLASS_NAMES)


def get_train_predict_dfs():
    df_train = pd.DataFrame(
        {
            "roi_id": ["site1"] * 50,
            "label": range(1, 50 + 1),
            "annotations": [1, 2, 1, 1, 2] * 10,
            **{
                feature_name: np.random.randn(50)
                for feature_name in TABLE_FEATURE_NAMES
            },
        }
    )
    df_predict = pd.DataFrame(
        {
            "roi_id": ["site2"] * 50,
            "label": range(1, 50 + 1),
            # "annotations": [1, 2, 1, 1, 2] * 10,
            **{
                feature_name: np.random.randn(50)
                for feature_name in TABLE_FEATURE_NAMES
            },
        }
    )
    return df_train, df_predict


# TODO: Add a real test.
def test_prediction():
    df_train, df_predict = get_train_predict_dfs()
    df_predict_with_nans = df_predict.copy()
    df_predict_with_nans.loc[[5, 10, 23], "feature1"] = np.nan
    c = get_classifier()
    c.add_features(df_train)
    c.train()
    predictions = c.predict(df_predict_with_nans)
    return predictions


def get_df():
    n = 5
    return pd.DataFrame(
        {
            "roi_id": ["site1"] * n,
            "label": range(1, n + 1),
            "annotations": [1, 2, 1, 1, 2],
            **TABLE_FEATURES,
        }
    )


def get_df_index_set():
    return get_df().set_index(INDEX_COLUMNS)


def get_df_with_changed_annotations():
    df_with_changed_annotations = get_df()
    df_with_changed_annotations.loc[
        df_with_changed_annotations.index[:3], "annotations"
    ] = [2, 1, 2]
    return df_with_changed_annotations


def get_df_nan_in_annotations():
    df_nan_in_annotations = get_df()
    df_nan_in_annotations.loc[[1, 4], "annotations"] = np.nan
    df_nan_in_annotations.loc[:, "roi_id"] = "site2"
    return df_nan_in_annotations


def get_df_nan_in_features():
    df_nan_in_features = get_df()
    df_nan_in_features.loc[[0, 1, 3], "feature1"] = np.nan
    df_nan_in_features.loc[:, "roi_id"] = "site3"
    return df_nan_in_features


def get_df_with_columns_to_delete():
    df_with_columns_to_delete = get_df()
    df_with_columns_to_delete.loc[
        df_with_columns_to_delete.index[:3], "annotations"
    ] = -1
    return df_with_columns_to_delete


def test_can_be_initialized_and_is_empty_upon_init():
    c = get_classifier()
    assert len(c._data) == 0


@pytest.mark.parametrize("df", [get_df(), get_df_index_set()])
def test_can_add_feature_table(df):
    c = get_classifier()
    df = get_df()
    c.add_features(df)
    assert len(c._data) == len(df)
    assert all(feature in c._data for feature in CLASSIFIER_FEATURE_NAMES)
    assert not any(feature in c._data for feature in DROP_COLUMNS)
    assert "hash" in c._data


def test_adding_the_same_table_multiple_times_has_no_effect():
    c1 = get_classifier()
    c2 = get_classifier()
    df = get_df()
    df_index = get_df_index_set()

    c1.add_features(df)
    c2.add_features(df_index)
    assert np.all(c1._data == c2._data)

    for _ in range(5):
        c2.add_features(df)
    assert np.all(c1._data == c2._data)


def test_rows_with_nan_in_annotation_not_added():
    c = get_classifier()
    df_nan = get_df_nan_in_annotations()

    c.add_features(df_nan)
    assert len(c._data == 3)
    assert np.all(c._data.annotations == c._data.annotations.dropna())


def test_rows_with_nan_in_features_not_added():
    c = get_classifier()
    df_nan = get_df_nan_in_features()

    c.add_features(df_nan)
    assert len(c._data) == 2
    assert np.all(c._data.feature1 == c._data.feature1.dropna())


def test_can_change_annotations():
    c = get_classifier()
    df = get_df()
    df2 = get_df_with_changed_annotations()
    assert np.any(df.annotations != df2.annotations)

    c.add_features(df)
    assert np.all(c._data.annotations.values == df.annotations.values)
    c.add_features(df2)
    assert np.all(c._data.annotations.values == df2.annotations.values)


def test_can_delete_annotations():
    c = get_classifier()
    df = get_df()
    df2 = get_df_with_columns_to_delete()

    c.add_features(df)
    assert len(c._data) == 5
    c.add_features(df2)
    assert len(c._data) == 2


def test_add_multi_site_with_changes_and_deletes():
    c = get_classifier()
    df = get_df()
    df_nan = get_df_nan_in_annotations()
    df_nan_feat = get_df_nan_in_features()
    df_change = get_df_with_changed_annotations()
    df_delete = get_df_with_columns_to_delete()

    c.add_features(df)
    assert len(c._data) == 5
    c.add_features(df_nan)
    assert len(c._data) == 8
    c.add_features(df_nan_feat)
    assert len(c._data) == 10
    c.add_features(df_change)
    assert len(c._data) == 10
    c.add_features(df_delete)
    assert len(c._data) == 7


def test_nans_in_non_classifier_features_have_no_effect():
    c = get_classifier()
    c2 = get_classifier()
    df = get_df()
    df_nans = df.copy()
    df_nans.loc[[1, 3, 4], "feature6"] = np.nan

    c.add_features(df)
    c2.add_features(df_nans)
    assert np.all(c._data == c2._data)
